;;;-*-Mode:LISP; Package: CHAOS; Base:10; Syntax:Common-lisp -*-
;;; $Id: op-check.lisp,v 1.1.1.1 1998/05/16 11:27:45 sawada Exp $
(in-package :chaos)
#|=============================================================================
				  System:CHAOS
				  Module:tools
			       File:op-check.lisp
=============================================================================|#
#-:chaos-debug
(declaim (optimize (speed 3) (safety 0) #-GCL (debug 0)))
#+:chaos-debug
(declaim (optimize (speed 1) (safety 3) #-GCL (debug 3)))

;;;*****************************************************************************
;;; CHECK TOOLS FOR OPERATORS
;;;*****************************************************************************

;;; **************************
;;; LAZYNESS(STRICTNESS) CHECK
;;; **************************
;;; CHECK-METHOD-STRINCTNESS method : -> Bool
;;;
;;; 1. constants(operations free from axioms) are always strict.
;;; 
(defun check-method-strictness (meth &optional
				     (module (or *last-module*
						 *current-module*))
				     report)
				     
  (unless module
    (with-output-chaos-error ('no-cntext)
      (princ "checking lazyness: no context module is specified!")
      ))
  ;;
  (with-in-module (module)
    (cond ((and (null (method-rules-with-different-top meth))
		(rule-ring-is-empty (method-rules-with-same-top meth)))
	   ;; the method has no rewrite rules
	   (if report
	       (show-method-strictness-report meth
					      (butlast (the-default-strategy
							(length (method-arity meth))))
					      nil)
	       (values (butlast (the-default-strategy (length (method-arity meth))))
		       nil)))
	  ;; the method has some rewrite rules associated with it.
	  ((or
	    ;; has some equational theory
	    (not (theory-is-empty-for-matching (method-theory meth)))
	    ;; the method is not free constructor.
	    (null (method-rules-with-different-top meth))
	    ;; has rules with different top and constant
	    ;; --> non-free constructor
	    (null (method-arity meth)))
	   ;; then the strategy is bottom up:
	   (if report
	       (show-method-strictness-report
		meth
		(butlast (the-default-strategy
			  (length (method-arity meth))))
		nil)
	       (values (butlast (the-default-strategy (length (method-arity meth))))
		       nil)))
	  ;;
	  (t
	   ;; the real work begins here.
	   (let ((l-ar (length (method-arity meth)))
		 (strategy nil)
		 (end-strategy nil))
	     (do ((occ 0 (1+ occ)))
		 ((<= l-ar occ))
	       (block is-variable
		 (let ((rr  (method-rules-with-same-top meth)))
		   (do ((rule (initialize-rule-ring rr) (rule-ring-next rr)))
		       ((end-of-rule-ring rr))
		     (unless (term-is-variable?
			      (term-arg-n (rule-lhs rule) occ))
		       ;; we eagarly evaluate non-variable argument.
		       (push (1+ occ) strategy)
		       ;; check next argument
		       (return-from is-variable))))
		 ;; we come here iff
		 ;; the arguments in lhs of all rules-with-same-top are
		 ;; variables or no rules-with-same-top.
		 (dolist (rule (method-rules-with-different-top meth))
		   (When
		       (let ((argn (term-arg-n (rule-lhs rule) occ))
			     (m (term-head (rule-lhs rule))))
			 (or (not (term-is-variable? argn))
			     ;; argn is not a variable.
			     (rule-is-builtin rule)
			     (not (and
				   ;; method is maximal?
				   ;; overloaded is necessarily a superset
				   ;; of lower. Note, overloaded always include
				   ;; self + default method, but lower not.
				   ;; thus #lower = #overloaded - 2 means the
				   ;; method is maximal.
				   (= (length (method-lower-methods meth))
				      (- (length (method-overloaded-methods meth)) 2))
				   (sort<= (nth occ (method-arity m))
					   (term-sort argn))
				   ))))
		     ;; method of lhs is not maximal,
		     ;; eagarly evaluates the non-variable argument.
		     (push (1+ occ) strategy)
		     ;; check the next arg.
		     (return-from is-variable)))
		 ;; come here if the occ-th argument is a variable,or
		 ;; method is maximal. can delay the evaluation.
		 (push (1+ occ) end-strategy)
		 ))
	     ;;
	     (if report
		 (show-method-strictness-report meth
						(reverse strategy)
						(reverse end-strategy))
		 (values (reverse strategy) (reverse end-strategy)))))
	  )))

(defun check-operator-strictness (op &optional (module (or *last-module*
							    *current-module*))
				     report)
  (unless module
    (with-output-chaos-error ('no-context)
      (princ "no context (current) module is given!")
      ))
  ;;
  (let* ((opinfo (if (opinfo-p op)
		     (prog1 op (setq op (opinfo-operator op)))
		     (get-operator-info op (module-all-operators module))))
	 (methods (opinfo-methods opinfo))
	 (res nil))
    (when report
      (print-next)
      (let ((*print-line-limit* 60))
	(format t "~%------------------------------------------------------------")
	(print-next)
	(print-centering
	 (format nil "* laziness of operator: ~a *" (operator-symbol op)))))
    ;;
    (dolist (meth methods)
      (unless (method-is-error-method meth)
	(multiple-value-bind (str-list1 str-list2)
	    (check-method-strictness meth module)
	  (when report
	    (with-in-module (module)
	      (print-next)
	      (format t "------------------------------------------------------------")
	      (show-method-strictness-report meth
					     str-list1
					     str-list2)))
	  (push (list meth str-list1 str-list2) res))))
    (nreverse res)))

(defun check-operator-strictness-whole (&optional (module (or *last-module*
							      *current-module*))
						  report)
  (unless module
    (with-output-chaos-error ('no-context)
      (princ "no context (current) module is specified!")
      ))
  ;;
  (let ((result nil))
    (dolist (opinfo (ops-to-be-shown module))
      (let ((op (opinfo-operator opinfo))
	    (res nil))
	(setq res (check-operator-strictness op module report))
	(push (cons op res) result)))
    (nreverse result)))
      
(defun show-method-strictness-report (method sl1 sl2)
  (print-next)
  (print-chaos-object method)
  (let ((*print-indent* (+ 2 *print-indent*))
	(strategy (or (method-rewrite-strategy method)
		      (method-supplied-strategy method)
		      (operator-strategy (method-operator method)))))
    (cond ((and (null sl1) (null sl2))
	   (print-next)
	   (princ "* NON-STRICT (lazy)."))
	  (t (when sl1
	       (print-next)
	       (format t "* strict on the arguments : ~{~^ ~a~}" sl1))
	     (when sl2
	       (print-next)
	       (format t "* may delay the evaluation on the arguments : ~{~^ ~a~}" sl2))))
    ;;
    (print-next)
    (if strategy
	(format t "- rewrite strategy:~{~^ ~a~}" strategy)
	(princ "- rewrite strategy: none"))
    (let ((axioms (method-all-rules method)))
      (when axioms
	(print-next)
	(princ "- axioms:")
	(let ((*print-indent* (+ 2 *print-indent*)))
	  (dolist (rl axioms)
	    (print-next)
	    (print-axiom-brief rl)))))))

;;; EOF
